@@ -356,10 +356,10 @@ static const char *operstate2str(unsigned char operstate)
 	return "";
 }
 
-static void extract_link(struct ifinfomsg *msg, int bytes,
+static connman_bool_t extract_link(struct ifinfomsg *msg, int bytes,
 				struct ether_addr *address, const char **ifname,
 				unsigned int *mtu, unsigned char *operstate,
-						struct rtnl_link_stats *stats)
+				struct rtnl_link_stats *stats)
 {
 	struct rtattr *attr;
 
@@ -389,8 +389,12 @@ static void extract_link(struct ifinfomsg *msg, int bytes,
 			break;
 		case IFLA_LINKMODE:
 			break;
+		case IFLA_WIRELESS:
+			return FALSE;
 		}
 	}
+
+	return TRUE;
 }
 
 static void process_newlink(unsigned short type, int index, unsigned flags,
@@ -407,7 +411,9 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 	GSList *list;
 
 	memset(&stats, 0, sizeof(stats));
-	extract_link(msg, bytes, &address, &ifname, &mtu, &operstate, &stats);
+	if (extract_link(msg, bytes, &address, &ifname, &mtu, &operstate,
+					&stats) == FALSE)
+		return;
 
 	snprintf(ident, 13, "%02x%02x%02x%02x%02x%02x",
 						address.ether_addr_octet[0],
@@ -489,7 +495,9 @@ static void process_dellink(unsigned short type, int index, unsigned flags,
 	GSList *list;
 
 	memset(&stats, 0, sizeof(stats));
-	extract_link(msg, bytes, NULL, &ifname, NULL, &operstate, &stats);
+	if (extract_link(msg, bytes, NULL, &ifname, NULL, &operstate,
+					&stats) == FALSE)
+		return;
 
 	if (operstate != 0xff)
 		connman_info("%s {dellink} index %d operstate %u <%s>",
@@ -603,11 +611,7 @@ static void process_newaddr(unsigned char family, unsigned char prefixlen,
 		 * have now properly configured interface with proper
 		 * autoconfigured address.
 		 */
-		char *interface = connman_inet_ifname(index);
-
-		__connman_resolver_redo_servers(interface);
-
-		g_free(interface);
+		__connman_resolver_redo_servers(index);
 	}
 }
 
@@ -1187,7 +1191,7 @@ static void rtnl_newnduseropt(struct nlmsghdr *hdr)
 	struct in6_addr *servers = NULL;
 	int i, nr_servers = 0;
 	int msglen = msg->nduseropt_opts_len;
-	char *interface;
+	int index;
 
 	DBG("family %d index %d len %d type %d code %d",
 		msg->nduseropt_family, msg->nduseropt_ifindex,
@@ -1199,8 +1203,8 @@ static void rtnl_newnduseropt(struct nlmsghdr *hdr)
 			msg->nduseropt_icmp_code != 0)
 		return;
 
-	interface = connman_inet_ifname(msg->nduseropt_ifindex);
-	if (!interface)
+	index = msg->nduseropt_ifindex;
+	if (index < 0)
 		return;
 
 	for (opt = (void *)&msg[1];
@@ -1221,7 +1225,7 @@ static void rtnl_newnduseropt(struct nlmsghdr *hdr)
 								sizeof(buf)))
 					continue;
 
-				connman_resolver_append_lifetime(interface,
+				connman_resolver_append_lifetime(index,
 							NULL, buf, lifetime);
 			}
 
@@ -1230,13 +1234,12 @@ static void rtnl_newnduseropt(struct nlmsghdr *hdr)
 
 			domains = rtnl_nd_opt_dnssl(opt, &lifetime);
 			for (i = 0; domains != NULL && domains[i] != NULL; i++)
-				connman_resolver_append_lifetime(interface,
+				connman_resolver_append_lifetime(index,
 						domains[i], NULL, lifetime);
 		}
 	}
 
 	g_free(domains);
-	g_free(interface);
 }
 
 static const char *type2string(uint16_t type)
